#######################################################################################################
# BINIAX
# Python version 1.0
# Original gameplay and concept by Jordan Tuzsuzov - (C) 25.04.2005
#
# GAME RULES :
# You have (for example 5x7) board with pairs of symbols (4 types or more).
# You have a symbol and you can move it in the following way :
# - You can move in all four direction in the grid, if there aren't pairs;
# - You can move over a pair if your symbol is the same as one of the symbols in the pair. In this
# case, your symbols becomes the other one from the pair, and the pair itself disapears.
# On every few seconds (depending the level) the field is scrolling down and a new row of pairs
# appears.
# The game ends, when you can not stay on the board.
#
# PRECONDITIONS :
# To run the Python version of the game under Linux/Unix, Windows, OSX or BeOS, you should install
# - Python 2.2 or higher from www.python.org
# - Pygame corresponding to your Python version from www.pygame.org
#
# COPYRIGHTS :
# This is a freeware gaming program, but even the source is available, the gameplay is copyrighted !
# As you can see, the code of this version is nothing special :)
# The official place of the game in Internet is www.biniax.com
#
# (C) 2005 - Jordan Tuzsuzov, www.tuzsuzov.com
#######################################################################################################

import math
import random
import os
import time
import pickle

import pygame
from pygame.locals import *

# GLOBALS. PARAMETRISATION OF THE GAME
# GAMEPLAY
_GRID_X = 5
_GRID_Y = 7

#GRAPHICS
_X1OFS = 8
_X2OFS = 58
_XOFS = 32
_YOFS = 0

_KNIGHT_XSIZE = 128
_KNIGHT_YSIZE = 64

_SCREEN_X = 640
_SCREEN_Y = 480

_SCROLL_SPEED = 70
_VIBRATE_TIME = 9
_GAME_SPEED = 60
_BLINK_SPEED = 10

#######################################################################################################
# CODE
#######################################################################################################

class Sprite :
    """Very simple sprite class"""
    def loadSprite( self, name ) :
        """Loads the image of the sprite"""
        pathname = os.path.join( "gfx", name )
        imag = pygame.image.load( pathname )
        self.image = imag.convert()
        self.height = self.image.get_height()
        self.width = self.image.get_width()
        
    def show( self, screen, x, y ) :
        """Shows the sprite with upper-left corner coordinates"""
        screen.blit( self.image, (x, y) )

    def showCentered( self, screen, x, y ) :
        """Shows the sprite with center coordinates"""
        screen.blit( self.image, (x - self.width/2, y - self.height/2) )

class Sound :
    """Simple sound class"""
    def loadSound( self, name ) :
        pathname = os.path.join( "sfx", name )
        self.sound = pygame.mixer.Sound( pathname )
        
    def play( self ) :
        self.sound.play()
        
class Knight :
    def __init__( self, maxCoat ) :
        self.coat1 = random.randint( 0, maxCoat )
        self.coat2 = random.randint( 0, maxCoat )
        while ( self.coat2 == self.coat1 ) :
            self.coat2 = random.randint( 0, maxCoat )
        self.status = True
        
    def setStatus( self, stat ) :
        self.status = stat
        
    def isHere( self ) :
        return self.status
        
    

class BiniaxGame :
    """Game class of Biniax game"""
    
    def __init__( self ) :
        """Default constructor"""
        pygame.init()
        pygame.mixer.init()
        self.screen = pygame.display.set_mode( (_SCREEN_X, _SCREEN_Y), 0, 16 )
        pygame.display.set_caption( "biniax" )
        self.initGrid()
        self.running = True
        self.gameMode = 0
        self.cursorBlink = 0
        self.vibrateX = 0
        
        self.font = pygame.font.Font( None, 32 )

        self.coats = [Sprite(),Sprite(),Sprite(),Sprite()]
        for i in range(0,4) :
            self.coats[ i ].loadSprite( "coat"+str(i)+".png" )
            
        self.field = Sprite()
        self.field.loadSprite( "board.png" )

        self.splash = Sprite()
        self.splash.loadSprite( "splash.png" )

        self.pausedGfx = Sprite()
        self.pausedGfx.loadSprite( "pause.png" )
        
        self.gameoverGfx = Sprite()
        self.gameoverGfx.loadSprite( "gameover.png" )

        self.bestscoreGfx = Sprite()
        self.bestscoreGfx.loadSprite( "bestscore.png" )

        self.knight = Sprite()
        self.knight.loadSprite( "knight.png" )
        
        self.umlot = [ Sprite(), Sprite() ]
        self.umlot[0].loadSprite( "umlot1.png" )
        self.umlot[1].loadSprite( "umlot2.png" )

        self.takeSfx, self.cannotSfx, self.scrollSfx, self.vibrateSfx = Sound(), Sound(), Sound(), Sound()
        self.takeSfx.loadSound( "sfx1.wav" )
        self.cannotSfx.loadSound( "sfx2.wav" )
        self.scrollSfx.loadSound( "sfx3.wav" )
        self.vibrateSfx.loadSound( "sfx4.wav" )
        
        self.newGame()

    def newGame( self ) :
        """New game"""
        self.initGrid()
        self.running = True
        self.cursorBlink = 0
        self.score = 0
        self.pause = False
        self.gameInit()
        self.speed = _SCROLL_SPEED
        self.loadHighscore()
        
    def saveHighscore( self ) :
        f = open( "biniax.hi", "w" )
        pickle.dump( self.bestScore, f )
        f.close()

    def loadHighscore( self ) :
        try:
            f = open( "biniax.hi", "r" )
            try:
                self.bestScore = pickle.load( f )
            finally:
                f.close()
        except IOError:
            self.bestScore = 0

    def initGrid( self ) :
        """Init the town for the first time"""
        
        # Player data
        self.playerX = self.playerY = 0
        self.playerCoat = 0
        
        # Grid data
        self.grid = []
        for y in range( 0, _GRID_Y ) :
            list1 = []
            for x in range( 0, _GRID_X ) :
                kni = Knight( 3 )
                kni.setStatus( False )
                list1.append( kni )
            self.grid.append( list1 )
        self.scrollDown( 3 )
        self.scrollDown( 3 )
        self.scrollDown( 2 )

    def scrollDown( self, level, play = False ) :
        """Moves the army of knights and adds a new row of knights at the end"""
        # Removes the first row
        self.grid.pop( 0 )
        # Move all to down, by adding a new row at the end
        list1 = []
        for x in range( 0, _GRID_X ) :
            list1.append( Knight( level ) )
        list1[ random.randint( 0, _GRID_X-1 ) ].setStatus( False )
        self.grid.append( list1 )
        # Move the player down, if can not take the front knight
        if self.playerY > 0 or not self.canDisable( self.playerX, self.playerY, self.playerCoat ) :
            self.playerY = self.playerY - 1
        if play :
            self.scrollSfx.play()
            

    def drawGrid( self ) :
        """View battle grid"""
        ypos = _SCREEN_Y - _KNIGHT_YSIZE
        
        for row in self.grid :
            xpos = 0
            for knight in row :
                if knight.isHere() :
                    self.coats[ knight.coat1 ].show( self.screen, self.vibrateX + xpos + _X1OFS, ypos + _YOFS )
                    self.coats[ knight.coat2 ].show( self.screen, self.vibrateX + xpos + _X2OFS, ypos + _YOFS )
                    self.knight.show( self.screen, self.vibrateX + xpos, ypos )
                xpos = xpos + _KNIGHT_XSIZE
                
            ypos = ypos - _KNIGHT_YSIZE
            
        xpos = self.playerX * _KNIGHT_XSIZE
        ypos = _SCREEN_Y - (self.playerY+1) * _KNIGHT_YSIZE
        self.coats[ self.playerCoat ].show( self.screen, xpos + _XOFS, ypos + _YOFS )
        if self.cursorBlink > _BLINK_SPEED / 2 :
            self.umlot[0].show( self.screen, xpos, ypos )
        else :
            self.umlot[1].show( self.screen, xpos, ypos )
        
    def canDisable( self, posx, posy, coat ) :
        if posy < 0 :
            return False
            
        if self.grid[ posy ][ posx ].isHere() :
            if self.grid[ posy ][ posx ].coat1 == coat :
                self.playerCoat = self.grid[ posy ][ posx ].coat2
                self.grid[ posy ][ posx ].setStatus( False )
                self.score = self.score + 10
                self.correctSpeed()
                self.takeSfx.play()
                return True
            elif self.grid[ posy ][ posx ].coat2 == coat :
                self.playerCoat = self.grid[ posy ][ posx ].coat1
                self.grid[ posy ][ posx ].setStatus( False )
                self.score = self.score + 10
                self.correctSpeed()
                self.takeSfx.play()
                return True
            else :
                self.cannotSfx.play()
                return False
        
        return True
                
    def moveLeft( self ) :
        if self.playerY < 0 :
            return
        if self.playerX > 0 :
            if not self.grid[ self.playerY ][ self.playerX - 1 ].isHere() :
                self.playerX = self.playerX - 1
            elif self.canDisable( self.playerX - 1, self.playerY, self.playerCoat ) :
                self.playerX = self.playerX - 1
                
            
    def moveRight( self ) :
        if self.playerY < 0 :
            return
        if self.playerX < _GRID_X-1 :
            if not self.grid[ self.playerY ][ self.playerX + 1 ].isHere() :
                self.playerX = self.playerX + 1
            elif self.canDisable( self.playerX + 1, self.playerY, self.playerCoat ) :
                self.playerX = self.playerX + 1

    def moveUp( self ) :
        if self.playerY < 0 :
            return
        if self.playerY < _GRID_Y-1 :
            if not self.grid[ self.playerY + 1 ][ self.playerX ].isHere() :
                self.playerY = self.playerY + 1
            elif self.canDisable( self.playerX, self.playerY + 1, self.playerCoat ) :
                self.playerY = self.playerY + 1

    def moveDown( self ) :
        if self.playerY < 0 :
            return
        if self.playerY > 0 :
            if not self.grid[ self.playerY - 1 ][ self.playerX ].isHere() :
                self.playerY = self.playerY - 1
            elif self.canDisable( self.playerX, self.playerY - 1, self.playerCoat ) :
                self.playerY = self.playerY - 1
    
    def gameInit( self ) :
        """Init players"""
        
    def correctSpeed( self ) :
        if self.speed > 20 and self.score % 150 == 0 :
            self.speed = self.speed - 1

    def start( self ) :
        """Game entry point"""
        
        # Main Game Cicle
        self.timer = pygame.time.Clock()
        waiter = self.speed
        scoreSaved = False
        while self.running :
            self.startFrame = pygame.time.get_ticks()
            
            # Draw field / splash
            if self.gameMode == 0 :
                self.splash.show( self.screen, 0, 0 )
            else :
                self.field.show( self.screen, 0, 0 )
                self.drawGrid()
                if self.playerY < 0 :
                    if self.score >= self.bestScore :
                        self.bestscoreGfx.show( self.screen, 64, 64 )
                        if not scoreSaved :
                            self.bestScore = self.score
                            self.saveHighscore()
                            scoreSaved = True
                    else :
                        self.gameoverGfx.show( self.screen, 64, 64 )
                if self.pause :
                    self.pausedGfx.show( self.screen, 64, 64 )
            
            textGfx = self.font.render( "SCORE : "+str(self.score), 1, (0,0,0) )
            self.screen.blit( textGfx, (3, 0) )
            textGfx = self.font.render( "BEST : "+str(self.bestScore), 1, (0,0,0) )
            self.screen.blit( textGfx, (290, 0) )
            
            pygame.display.flip()
            
            # Event processing
            for event in pygame.event.get():
                if event.type == QUIT:
                    self.running = False
                elif event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        if self.gameMode == 1 :
                            self.gameMode = 0
                        elif self.gameMode == 0 :
                            self.running = False
                    elif event.key == K_LEFT :
                        self.moveLeft()
                    elif event.key == K_RIGHT :
                        self.moveRight()
                    elif event.key == K_UP :
                        self.moveUp()
                    elif event.key == K_DOWN :
                        self.moveDown()
                    elif event.key == K_SPACE :
                        if self.gameMode == 1 and self.playerY < 0 :
                            self.gameMode = 0
                        elif self.gameMode == 1 and self.playerY >= 0 :
                            self.pause = not self.pause
                        elif self.gameMode == 0 :
                            self.newGame()
                            waiter = self.speed
                            self.gameMode = 1
                            scoreSaved = False
                        
            if self.gameMode == 1 and not self.pause :
                if waiter <= 0 :
                    self.scrollDown( 3, True )
                    waiter = self.speed
                    self.vibrateX = 0
                elif waiter == _VIBRATE_TIME :
                    self.vibrateSfx.play()
                elif waiter < _VIBRATE_TIME :
                    self.vibrateX = ((waiter % 3) - 1) * 2
                waiter = waiter - 1
            
                self.cursorBlink = (self.cursorBlink + 1) % 10
            
            
            # Synchronize with timer
            while ( pygame.time.get_ticks()-self.startFrame ) < _GAME_SPEED : time.sleep( 0 )

        # End of main game cicle
        
# START THE APPLICATION
print "Starting : BINIAX"
print "Gameplay version 1.0, Software version 1.0.py"
print "Original concept and gameplay by Jordan Tuzsuzov, (C) 2005"
print "For more information visit : http://www.tuzsuzov.com"

BiniaxGame().start()
